#!/bin/python3
import os
import pandas as pd
from Bio.Seq import Seq
from Bio.SeqUtils import MeltingTemp
from Bio import SeqIO
import argparse
import itertools
import json
import requests
import primer3

# Define the arguments to be parsed
parser = argparse.ArgumentParser(description='PCR parameter inputs')
parser.add_argument('--target_tm', type=int, help='Annealing temperature (in Celsius)', default = 55.0)
parser.add_argument('--salt_concentration', type=float, help='Salt concentration (in nM)', default = 50)
parser.add_argument('--product_size_max', type=float, help='maximum length of PCR products in nucleotides', default = 2000)
parser.add_argument('--req_five', type=str, help="Require the 5' end of the primer to bind?", default = True)
parser.add_argument('--primer_seq', type=str, help="Primer sequences, one per line", required = True)
parser.add_argument('--input_file', type=str, help="Reference FASTA file", required = True)
parser.add_argument('--output', type=str, help="Output file name", default = "in_silico_PCR")
parser.add_argument('--Q5', action='store_true', help='Whether to use Q5 approximation settings for Tm calculations.', default=False)

# check if blastn is in the system path
try:
    status = os.system('which blastn > /dev/null 2>&1')
    if status != 0:
        print('Error: BLAST is not installed.')
        exit(1)
except OSError as e:
    print('Error:', e)
    exit(1)

# Check if required arguments are missing and print help message
args = parser.parse_args()
if not args.primer_seq or not args.input_file:
    parser.print_help()
    exit()

# Define the inputs
annealing_temp = args.target_tm - 4
salt_conc = args.salt_concentration
max_amplicon_len = args.product_size_max
req_five = args.req_five
primer_seq = args.primer_seq
ref_fasta_file = args.input_file
out_file = args.output


def q5_melting_temp(seq1, seq2, salt=0.5):
    url = "https://tmapi.neb.com/tm/q5/%s/%s/%s?&fmt=short" % (salt, seq1, seq2)
    response = requests.get(url)
    json_string = response.json()
    tm1 = json_string["data"]["tm1"]
    tm2 = json_string["data"]["tm2"]
    return [tm1, tm2]

def calc_tm(seq, seq2='', Q5=args.Q5):
    if Q5:
        # tm = MeltingTemp.Tm_NN(seq, c_seq = seq2, nn_table=MeltingTemp.DNA_NN4, dnac1=2500, dnac2=2500, Na=40, K=0, Tris=0, Mg=0.4, dNTPs=0, saltcorr=5.0)
        # tm = q5_melting_temp(str(seq), str(seq2), salt_conc/1000)
        if seq2 != '':
            tm = primer3.bindings.calcTm(seq, dv_conc=2, mv_conc=70, dna_conc=3300)
        else:
            primer3.calcHeterodimer(seq, seq2, temp_c=100, dv_conc = 2, mv_conc = 70, dna_conc = 3300)
    else:
        tm = MeltingTemp.Tm_NN(seq, c_seq = seq2, Na=salt_conc, dnac1=250, dnac2=0, saltcorr=7, nn_table=MeltingTemp.DNA_NN4, selfcomp=False, check=True, shift=0.0)
    return tm

def find_binding_positions(primer_seq, ref_fasta_file, annealing_temp, req_five, salt_conc):
    # Run BLAST to search for primer sequence against reference FASTA
    blast_cmd = f"blastn -query {primer_seq} -subject {ref_fasta_file} -task blastn-short -outfmt '6 qseqid sseqid qstart qend sstart send qseq sseq mismatch length' > blast_output.txt"
    os.system(blast_cmd)
    
    blast_df = pd.DataFrame(columns=['Query ID', 'Subject ID', 'Query Start', 'Query End', 'Subject Start', 'Subject End', 'Query Sequence Match', 'Direction', 'Binding Position', 'Mismatches', 'Binding Length'])
    
    with open('blast_output.txt', 'r') as f:
        for line in f:
            fields = line.strip().split('\t')
            qseqid, sseqid, qstart, qend, sstart, send, qseq, sseq, mismatch, length = fields
            direction = '+' if int(sstart) < int(send) else '-'
            binding_pos = sstart if direction == '+' else send
            # Check if primer binds at the specified annealing temperature
            if req_five and int(qstart) >= 2:
                continue
            # tm = MeltingTemp.Tm_NN(Seq(sseq), Na=salt_conc, dnac1=250, dnac2=0, saltcorr=7, nn_table=MeltingTemp.DNA_NN4, selfcomp=False, check=True, shift=0.0)
            tm = calc_tm(Seq(sseq))
            if annealing_temp <= tm:
                blast_df.loc[len(blast_df)] = [qseqid, sseqid, qstart, qend, sstart, send, sseq, direction, binding_pos, mismatch, length]
    return blast_df

def find_compatible_pairs(blast_df, max_len):
    pairs = itertools.combinations(blast_df.index, 2)
    compatible_pairs = []
    for pair in pairs:
        row1 = blast_df.loc[pair[0]]
        row2 = blast_df.loc[pair[1]]
        amp_size = int(row1['Binding Position']) - int(row2['Binding Position'])
        if abs(amp_size) <= max_len and row1['Direction'] != row2['Direction'] and row1['Subject ID'] == row2['Subject ID']:
            if row1['Direction'] == '+' and amp_size < 0:
                compatible_pairs.append({'qseq1': row1['Query Seq'], 'qseq1_input': row1['Sequence'], 'qstart1': row1['Query Start'], 'qend1': row1['Query End'], 'direction1': row1['Direction'], 'mismatch1': row1['Mismatches'],
                                         'qseq2': row2['Query Seq'], 'qseq2_input': row2['Sequence'], 'qstart2': row2['Query Start'], 'qend2': row2['Query End'], 'direction2': row2['Direction'], 'mismatch2': row2['Mismatches'],
                                         'binding_pos_diff': abs(amp_size), 'reference': row1['Subject ID'], 'ref_region': str(row2['Binding Position']) + ", " + str(row1['Binding Position'])})
            elif row1['Direction'] == '-' and amp_size > 0:
                compatible_pairs.append({'qseq1': row2['Query Seq'], 'qseq2_input': row2['Sequence'], 'qstart1': row2['Query Start'], 'qend1': row2['Query End'], 'direction1': row2['Direction'], 'mismatch2': row2['Mismatches'],
                                         'qseq2': row1['Query Seq'], 'qseq1_input': row1['Sequence'], 'qstart2': row1['Query Start'], 'qend2': row1['Query End'], 'direction2': row1['Direction'], 'mismatch1': row1['Mismatches'],
                                         'binding_pos_diff': abs(amp_size), 'reference': row1['Subject ID'], 'ref_region': str(row1['Binding Position']) + ", " + str(row2['Binding Position'])})
            else:
                continue
    return pd.DataFrame(compatible_pairs)

def find_oligo_dimers(fasta_file, temp, salt_conc):
    # Read in the fasta sequences
    seq_records = list(SeqIO.parse(fasta_file, "fasta"))

    # Create a list to hold the dimer pairs and melting temperatures
    dimers = []
    melting_temps = []

    # Loop through all possible pairs of sequences
    for seq1, seq2 in itertools.combinations(seq_records, 2):
        # Calculate the melting temperature of the dimer
        # dimer_tm = MeltingTemp.Tm_NN(str(seq1.seq + seq2.seq), nn_table=MeltingTemp.DNA_NN4, Na=salt_conc, saltcorr=7)
        dimer_tm = calc_tm(str(seq1.seq + seq2.seq))

        # Check if the dimer's melting temperature is above the given temperature
        if dimer_tm > temp:
            # Add the dimer's names and melting temperature to the lists
            dimers.append((seq1.id, seq2.id))
            melting_temps.append(round(dimer_tm))

    # Create a pandas dataframe from the dimer pairs and melting temperatures
    df = pd.DataFrame({'Sequence1': [d[0] for d in dimers],
                       'Sequence2': [d[1] for d in dimers],
                       'MeltingTemp': melting_temps})

    return df

def is_valid_dna_sequence(sequence):
    """Checks if a DNA sequence only contains valid characters."""
    valid_chars = set('ACGTN')
    return all(char in valid_chars for char in sequence.upper())

def main():
    i=0
    with open(primer_seq, 'r') as infile, open(primer_seq+'.fasta', 'w') as outfile:
        # Initialize a dictionary to keep track of sequence names
        seen_names = {}
        i+=1
        for line in infile:
            columns = line.strip().split('\t')
            if len(columns) == 1:
                # Remove any leading/trailing whitespace from the sequence
                sequence = line.strip()
                if not (is_valid_dna_sequence(sequence)):
                    print("error in sequence on: " + line)
                    break
                # Generate a unique sequence name
                name = f'sequence{i}'
            else:
                sequence, name = columns
                if not (is_valid_dna_sequence(sequence)):
                    print("error in sequence on: " + name)
                    break
            # Check if the name has already been used
            if name in seen_names:
                # Increment the count for this name
                count = seen_names[name] + 1
                # Generate a new name by appending the count to the original name
                new_name = f'{name}_{count}'
                # Update the dictionary with the new count
                seen_names[name] = count
                # Use the new name for the sequence
                name = new_name
            else:
                # Add the name to the dictionary with a count of 1
                seen_names[name] = 1
            # Write the sequence to the output file in FASTA format
            outfile.write(f'>{name}\n{sequence}\n')


    # primer dimers
    dimers = find_oligo_dimers(primer_seq+'.fasta', annealing_temp, salt_conc)
    # print("Writing output to " + out_file + '_primer_dimears.tsv')
    # dimers.to_csv(out_file + '_primer_dimers.tsv', index=False, sep="\t")


    blast_df = find_binding_positions(primer_seq+'.fasta', ref_fasta_file, annealing_temp, req_five, salt_conc)

    # Read in the FASTA file as a dictionary of SeqRecord objects
    fasta_dict = SeqIO.to_dict(SeqIO.parse(primer_seq+'.fasta', "fasta"))

    # Iterate through the rows of the DataFrame and replace the add primer with sequences
    for i, row in blast_df.iterrows():
        query_id = row["Query ID"]
        if query_id in fasta_dict:
            sequence = str(fasta_dict[query_id].seq)
            blast_df.at[i, "Sequence"] = sequence
        else:
            blast_df.at[i, "Sequence"] = ""

    blast_df = blast_df.rename(columns={'Query ID': 'Query Seq'})

    amp_df = find_compatible_pairs(blast_df, max_len=max_amplicon_len)
    # print("Writing output to " + out_file + '.tsv')
    # amp_df.to_csv(out_file + '.tsv', index=False, sep="\t")



    # get interactions
    amplicons = []
    ref_seq = SeqIO.read(ref_fasta_file, "fasta").seq
    for index, row in amp_df.iterrows():
        ref_region = row['ref_region']
        #add in the primer sequences
        if row['direction1'] == '-':
            pf_seq = Seq(row['qseq1_input'])
            pf_seq = str(pf_seq.reverse_complement())
        else:
            pf_seq = Seq(row['qseq1_input'])
        if row['direction2'] == '-':
            pr_seq = Seq(row['qseq2_input'])
            pr_seq = str(pf_seq.reverse_complement())
        else:
            pr_seq = Seq(row['qseq2_input'])
        ref_start, ref_end = ref_region.split(",")
        ref_end = int(ref_end)
        ref_start = int(ref_start)
        if ref_start > ref_end:
            amplicon_seq = ref_seq[ref_end-1:ref_start]
        else:
            amplicon_seq = ref_seq[ref_start-1:ref_end]
        amplicon_seq = str(pf_seq) + str(amplicon_seq) + str(pr_seq)
        amplicons.append({'Forward_primer':  row['qseq1'], 'Reverse_primer':  row['qseq2'], 'amplicon_seq': amplicon_seq})


    tm_data = []
    for pair in itertools.combinations(amplicons, 2):
        # tm = MeltingTemp.Tm_NN(pair[0]['amplicon_seq'], pair[1]['amplicon_seq'], Na=salt_conc, saltcorr=7)
        try:
            tm = calc_tm(seq=pair[0]['amplicon_seq'], seq2=pair[1]['amplicon_seq'])
            tm_data.append({"amplicon1_PF": pair[0]['Forward_primer'],
                "amplicon1_PR": pair[0]['Reverse_primer'],
                "amplicon2_PF": pair[1]['Forward_primer'],
                "amplicon2_PR": pair[1]['Reverse_primer'], 
                'tm': str(tm)})
        except:
            continue

    tm_df = pd.DataFrame(tm_data)
    # print("Writing output to " + out_file + '_amplicon_interactions.tsv')
    # tm_df.to_csv(out_file + '_amplicon_interactions.tsv', index=False, sep="\t")


    with pd.ExcelWriter(out_file + '.xlsx', engine='xlsxwriter') as writer:
        amp_df.to_excel(writer, sheet_name='amplicons', index=False)
        dimers.to_excel(writer, sheet_name='dimers', index=False)
        tm_df.to_excel(writer, sheet_name='interactions', index=False)

    # cleanup
    os.remove(primer_seq+'.fasta')
    os.remove("blast_output.txt")


if __name__ == '__main__':
    main()





#!/bin/python
from Bio import SeqIO
from Bio.SeqUtils import MeltingTemp as mt
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
import primer3
import pandas as pd
import itertools
import argparse
import numpy as np
import concurrent.futures
import sys
import random
import requests
from sklearn.cluster import AgglomerativeClustering
import tempfile
import subprocess
import shutil
import os
# import matplotlib.pyplot as plt

parser = argparse.ArgumentParser(description='PCR primer design')
# Define input arguments
# parser.add_argument('--cpus', type=int, help='number of threads (even though it says cpus)', default=1000)
parser.add_argument('--region_file', type=str, help='The path to the primer design region file. Two columns, start position and end position (1-based). tsv or xlxs', required=True)
parser.add_argument('--input_file', type=str, help="Reference FASTA file", required = True)
parser.add_argument('--target_tm', type=float, help='The desired melting temperature (Tm) for the primers.', default=65)
parser.add_argument('--primer_len', type=int, help='The desired length of the primers.', default=20)
parser.add_argument('--product_size_min', type=int, help='The desired min size for the PCR product. Default is 400.', default=400)
parser.add_argument('--product_size_max', type=int, help='The desired max size for the PCR product. Default is 800.', default=800)
parser.add_argument('--ret', type=int, help='The maximum number of primer pairs to return.', default=100)
parser.add_argument('--Q5', action='store_true', help='Whether to use Q5 approximation settings for Tm calculations.', default=True)
parser.add_argument('--background', type=str, help='The path to the mispriming library FASTA file.', default='')
parser.add_argument('--output', type=str, help='Output name.', default='MultiPlexPrimerSet')
parser.add_argument('--eval', type=int, help='The maximum number of primer sets to evaluate.', default=10000)
parser.add_argument('--wiggle', type=int, help='Half the region around the optimal Tm', default=3)
parser.add_argument('--ill_adapt', action='store_true', help='Add Illumina partial adapters', default=False)
parser.add_argument('--clamp', type=int, help='Require GC clamp', default=0)
parser.add_argument('--poly', type=int, help='Maximum allowable length of a mononucleotide repeat (poly-X) in the primer sequence', default=3)
args = parser.parse_args()

product_size_range = (args.product_size_min, args.product_size_max)

def design_primers(input_file, region_start, region_end, target_tm=args.target_tm, primer_len=20, product_size_range=(400, 800), name='', ret=100, Q5=True, background='', wiggle=args.wiggle, clamp=args.clamp, poly=args.poly):
    """
    This function takes a FASTA file, start and end positions of a target region, and desired melting temperature
    and primer length, and returns the best primer set for that region using Biopython and primer3.

    Args:
    input_file (str): The path to the input FASTA file.
    region_start (int): The start position of the target region (1-based).
    region_end (int): The end position of the target region (1-based).
    target_tm (float): The desired melting temperature (Tm) for the primers.
    primer_len (int): The desired length of the primers.
    product_size_range (tuple): The desired size range for the PCR product.
    name (str): The name of the primer set.
    ret (int): The maximum number of primer pairs to return.
    Q5 (bool): Whether to use Q5 settings for primer3.
    background (str): The name of a mispriming library to use for primer3.

    Returns:
    list: A list of dicts containing information about the primer pairs.
    """

    # Parse the input FASTA file and extract the target region sequence.
    # it seems that primer3 doesnt look for non-specificity in the input file, so need to append the target non-regions to the background fasta.
    print("Picking initial primers for " + name)
    # giving issues on some systems because of the -i flag
    # command = "sed -i '$d;$d' " + background
    # subprocess.run(command, shell=True)
    # Create a temporary file to hold the modified output
    # with tempfile.NamedTemporaryFile(mode='w', delete=False) as tmp_file:
    #     # Use sed to remove the last four lines from the input file and write the result to the temporary file
    #     subprocess.run(f"sed '$d;$d;$d;$d' '{background}'", stdout=tmp_file, shell=True) #doesnt seem to work on linux

    # # Move the temporary file over the original file
    # shutil.move(tmp_file.name, background)

    # Read the file and store its content
    with open(background, 'r') as file:
        lines = file.readlines()

    # Remove the last four lines
    new_lines = lines[:-4]

    # Write the updated content back to the file
    with open(background, 'w') as file:
        file.writelines(new_lines)

    if region_start - 10000 > 0:
        target_seq = input_file.seq[region_start-10000:region_end+10000]
        seq_target = (10000-1, region_end-region_start)
        add_to = 10000
        with open(background, "a") as fasta_file:
            fasta_file.write(">appended_sequence\n")
            fasta_file.write(str(input_file.seq[0:region_start-10000]) + "\n")
            fasta_file.write(">appended_sequence2\n")
            fasta_file.write(str(input_file.seq[region_end+10000:]))
    else:
        target_seq = input_file.seq[0:region_end+10000]
        seq_target = (region_start-1, region_end-region_start)
        add_to = 0
        with open(background, "a") as fasta_file:
            fasta_file.write(">appended_sequence\n")
            fasta_file.write(str(input_file.seq[region_end+10000:]) + "\n")
            fasta_file.write(">appended_sequence2\n")
            fasta_file.write("gagagagaga")


    # target_seq = input_file.seq
    # seq_target = (region_start-1, region_end-region_start+2)

    # shutil.copy(background, "/home/semiquant/Desktop/tmp/tmp.fasta")

    # Set up the primer3 input parameters.
    input_params = {
    'SEQUENCE_ID': input_file.id,
    'SEQUENCE_TEMPLATE': str(target_seq),
    'SEQUENCE_TARGET': seq_target,
    'PRIMER_OPT_SIZE': primer_len,
    'PRIMER_MIN_SIZE' : 10,
    'PRIMER_MAX_SIZE' : 36,
    'PRIMER_PICK_INTERNAL_OLIGO': 0,
    'PRIMER_MISPRIMING_LIBRARY': background,
    }

    global_args = {
    'PRIMER_MIN_GC': 40,
    'PRIMER_MAX_GC': 80,
    'PRIMER_NUM_RETURN': ret,
    'PRIMER_EXPLAIN_FLAG': 1,
    'PRIMER_MIN_TM': target_tm-wiggle,
    'PRIMER_OPT_TM': target_tm,
    'PRIMER_MAX_TM': target_tm+wiggle,
    'PRIMER_PRODUCT_SIZE_RANGE': product_size_range,
    'PRIMER_INTERNAL_MAX_POLY_X':poly,
    }
        
    if Q5:
        global_args.update({
        'PRIMER_SALT_DIVALENT': 2,
        'PRIMER_SALT_MONOVALENT': 70,
        'PRIMER_DNA_CONC': 3300,
        'PRIMER_TM_FORMULA': 1,
        'PRIMER_SALT_CORRECTIONS': 2,
        })

    if clamp > 0:
        global_args.update({'PRIMER_GC_CLAMP': clamp,})

    # Run primer3 to design the primers.
    primers = primer3.bindings.designPrimers(input_params, global_args)
    no_miss = False
    if 'PRIMER_PAIR_NUM_RETURNED' not in primers or primers['PRIMER_PAIR_NUM_RETURNED'] == 0:
        print("No primers found for: " + name + "; trying with relaxed restraints.")
        del global_args['PRIMER_INTERNAL_MAX_POLY_X']
        global_args.update({'PRIMER_PICK_ANYWAY': 1,})
        primers = primer3.bindings.designPrimers(input_params, global_args)
        if 'PRIMER_PAIR_NUM_RETURNED' not in primers or primers['PRIMER_PAIR_NUM_RETURNED'] == 0:
            del input_params['PRIMER_MISPRIMING_LIBRARY']
            global_args.update({'PRIMER_MIN_TM': target_tm - wiggle * 1.5,})
            global_args.update({'PRIMER_MAX_TM': target_tm + wiggle * 1.5,})
            primers = primer3.bindings.designPrimers(input_params, global_args)
            no_miss = True
            if 'PRIMER_PAIR_NUM_RETURNED' not in primers or primers['PRIMER_PAIR_NUM_RETURNED'] == 0:
                global_args.update({'PRIMER_MIN_TM': target_tm - wiggle * 5,})
                global_args.update({'PRIMER_MAX_TM': target_tm + wiggle * 5,})
                product_size_range_tmp = (50, round(product_size_range[1] * 1.5))
                global_args.update({'PRIMER_PRODUCT_SIZE_RANGE': product_size_range_tmp,})
                global_args.update({'PRIMER_MIN_SIZE': 5,})
                global_args.update({'PRIMER_MAX_SIZE': 36,})
                primers = primer3.bindings.designPrimers(input_params, global_args)
    try:
        # Check if any primer pairs were returned.
        if 'PRIMER_PAIR_NUM_RETURNED' not in primers or primers['PRIMER_PAIR_NUM_RETURNED'] == 0:
            return []
        else:
            if no_miss:
                print("No mispriming library used for " + name)
            # Extract information about the primer pairs.
            primer_pairs = []
            for i in range(primers['PRIMER_PAIR_NUM_RETURNED']):
                forward_seq = primers[f'PRIMER_LEFT_{i}_SEQUENCE']
                reverse_seq = primers[f'PRIMER_RIGHT_{i}_SEQUENCE']
                forward_tm = primers[f'PRIMER_RIGHT_{i}_TM']
                reverse_tm = primers[f'PRIMER_LEFT_{i}_TM']
                product_size = primers[f'PRIMER_PAIR_{i}_PRODUCT_SIZE']
                start = primers['PRIMER_LEFT_{}'.format(i)][0] + region_start - add_to
                end = primers['PRIMER_RIGHT_{}'.format(i)][0] + region_start - add_to

                # Check that the product size is within the specified range.
                # if product_size < int(product_size_range[0]) or product_size > int(product_size_range[1]):
                if args.ill_adapt:
                    # add illumina partial adapter, dont adjust Tm as these are tails
                    forward_seq = "ACACTCTTTCCCTACACGACGCTCTTCCGATCT" + forward_seq
                    reverse_seq = "GACTGGAGTTCAGACGTGTGCTCTTCCGATCT" + reverse_seq
                primer_pairs.append({
                    'Forward Primer': forward_seq,
                    'Reverse Primer': reverse_seq,
                    'Forward tm': forward_tm,
                    'Reverse tm': reverse_tm,
                    'Product Size': product_size,
                    'Binding Start': str(start),
                    'Binding End': str(end),
                    'name': name
                    })
    except Exception as e:
        print("An error occurred:", e)

    return primer_pairs

def q5_melting_temp(seq1, seq2="", salt=0.5):
    url = "https://tmapi.neb.com/tm/q5/%s/%s/%s?&fmt=short" % (salt, seq1, seq2)
    response = requests.get(url)
    json_string = response.json()
    tm1 = json_string["data"]["tm1"]
    # tm2 = json_string["data"]["tm2"]
    return tm1 #[tm1, tm2]

def is_valid_combination(combination, names):
    if set(entry['name'] for entry in combination) == names:
        return combination
    return []

def evaluate_combination(comb, Q5=args.Q5):
    # Create a list of primer pairs
    primer_pairs = [(p1, p2) for p1, p2 in itertools.combinations(comb, 2)]
    # Extract the primer sequences from the primer pairs
    primer_pairs_sequences = [(p1['Forward Primer'], p2['Reverse Primer']) for p1, p2 in primer_pairs]
    # Calculate heterodimer formation energy
    heterodimer_scores = []
    for p in primer_pairs_sequences:
        if Q5:
            heterodimer_scores.append(primer3.calcHeterodimer(p[0], p[1], temp_c=args.target_tm, dv_conc = 2, mv_conc = 70, dna_conc = 3300).dg) #gibbs free energy
        else:
            heterodimer_scores.append(primer3.calcHeterodimer(p[0], p[1], temp_c=args.target_tm).dg) #gibbs free energy
    product_size_weight=0.2
    heterodimer_score_weight =0.8
    score = {'size': product_size_range,
             'tmp': tm_range,
             'mean': np.mean(heterodimer_scores),
             'range': abs(np.min(heterodimer_scores)) - abs(np.max(heterodimer_scores))
             # 'weighted_score': product_size_weight * (1 - product_size_range) + heterodimer_score_weight * (1 - abs(np.mean(heterodimer_scores)) / 100)
            }    
    return score, comb


# Define the features to consider for clustering
def get_features(primer, target_tm=args.target_tm):
    fwd_tm = primer['Forward tm']
    rev_tm = primer['Reverse tm']
    avg_tm = (fwd_tm + rev_tm) / 2
    product_size = primer['Product Size']
    tm_difference = abs(fwd_tm - rev_tm)
    dimer_prob = primer3.calcHeterodimer(primer['Forward Primer'], primer['Reverse Primer'], temp_c=target_tm, dv_conc = 2, mv_conc = 70, dna_conc = 3300).dg
    return (avg_tm, product_size, tm_difference, dimer_prob)
    

def cluster_primers(primers, target_tm=args.target_tm, distance_threshold=10):
    # Cluster the primers based on their features
    X = [get_features(primer) for primer in primers]
    clusterer = AgglomerativeClustering(n_clusters=None, affinity='euclidean', linkage='ward', distance_threshold=distance_threshold)
    clusters = clusterer.fit_predict(X)

    # Select the largest cluster
    name_counts = {}
    for i, cluster in enumerate(clusters):
        name = primers[i]['name']
        if cluster in name_counts:
            name_counts[cluster].add(name)
        else:
            name_counts[cluster] = {name}

    most_unique_cluster = max(name_counts, key=lambda k: len(name_counts[k]))
    largest_cluster = [primers[i] for i, cluster in enumerate(clusters) if cluster == most_unique_cluster]

    # Add missing names from the next closest cluster until all names are included
    all_names = set(primer['name'] for primer in primers)
    selected_names = set(primer['name'] for primer in largest_cluster)
    missing_names = all_names - selected_names

    while missing_names:
        # Find the closest cluster that contains missing names
        closest_cluster_idx = None
        closest_distance = float('inf')
        for i, cluster in enumerate(clusters):
            if cluster == most_unique_cluster:
                continue
            cluster_names = set(primers[j]['name'] for j in range(len(primers)) if clusters[j] == cluster)
            common_names = cluster_names & missing_names
            if common_names:
                distance = np.linalg.norm(np.array(X[i]) - np.array(X[0]))
                if distance < closest_distance:
                    closest_cluster_idx = cluster
                    closest_distance = distance

        # Add the missing names to the largest cluster
        closest_cluster = [primers[i] for i, cluster in enumerate(clusters) if cluster == closest_cluster_idx]
        for primer in closest_cluster:
            if primer['name'] in missing_names:
                largest_cluster.append(primer)
                missing_names.remove(primer['name'])

    # Remove duplicates by name
    selected_primers = []
    selected_names = set()
    for primer in largest_cluster:
        name = primer['name']
        if name not in selected_names:
            selected_primers.append(primer)
            selected_names.add(name)
    return selected_primers


def plot_cluster(clusterer, primers):
    # Obtain the cluster labels
    labels = clusterer.labels_

    # Plot the clusters
    fig, ax = plt.subplots()
    ax.scatter([x[0] for x in X], [x[1] for x in X], c=labels, cmap='viridis')

    # Add labels for each point
    for i, primer in enumerate(primers):
        ax.annotate(primer['name'], (X[i][0], X[i][1]))

    # Customize the plot
    ax.set_xlabel('Average TM')
    ax.set_ylabel('Product Size')
    ax.set_title('Primer Clustering')

    # Obtain the cluster labels
    labels = clusterer.labels_

    # Create a dictionary of primer names and their cluster assignments
    cluster_dict = defaultdict(list)
    for i, primer in enumerate(primers):
        cluster_dict[primer['name']].append(labels[i])

    # Print the primer names and their cluster assignments as a table
    rows = []
    for name, clusters in cluster_dict.items():
        rows.append([name, ', '.join(str(c) for c in clusters)])
    
    return (fig, tabulate(rows, headers=['Primer Name', 'Clusters']))


def main():
    #########################################
    # Read the file into a pandas DataFrame #
    file_path = args.region_file  # or 'file.xlsx'
    if file_path.endswith('.tsv'):
        df = pd.read_csv(file_path, sep='\t')
    else:
        df = pd.read_excel(file_path, engine='openpyxl') 

    seen_names = set()
    modified_names = {}

    primers_all = []

    record = SeqIO.read(args.input_file, "fasta")


    # check if regions overlap
    # Sort the DataFrame by 'start' column
    df = df.sort_values(by='start')
    overlap = df.loc[(df['start'] < df['end'].shift()) & (df['end'] > df['start'].shift())]
    if not overlap.empty:
        print("Overlaps found, please edit input")
        print(overlap)
        exit(1)


    # Iterate over each row
    for index, row in df.iterrows():
        # Access the "start" and "end" values using the column names
        # name=str(row['name'])
        start = int(row['start'])
        end = int(row['end'])
        name = 'Target ' + str(start) + "-" + str(end)

        if end - start > int(product_size_range[1]):
            print("Region " + str(start) + ", ", str(end) + " larger than product_size_range!")
            pass

        # Check if name is already in seen_names
        if name in seen_names:
            # If so, modify the name to make it unique
            if name in modified_names:
                modified_name = modified_names[name]
            else:
                modified_name = name + '_' + str(random.randint(1, 100))
                modified_names[name] = modified_name
            # Update name in row
            df.at[index, 'name'] = modified_name
            seen_names.add(modified_name)
            name = modified_name
        else:
            seen_names.add(name)

        # copy background fasta
        tmp_dir = tempfile.TemporaryDirectory()
        tmp_back_n = tmp_dir.name
        if args.background != '':
            shutil.copy(args.background, tmp_back_n)
            tmp_back = os.path.join(tmp_back_n, os.path.basename(args.background))
        else:
            tmp_back = os.path.join(tmp_back_n, 'no_back.fasta')
            # with open(tmp_back, "a") as fasta_file:
            #     fasta_file.write(">random_testing\n")
            #     fasta_file.write("gcaggcaggcaggcag" + "\n")
            #     fasta_file.write(">random_testing2\n")
            #     fasta_file.write("gcaggcaggcaggccag" + "\n")
        #     # Create a temporary file for writing
        #     with tempfile.NamedTemporaryFile(delete=False) as tmp_back:
        #         # You can write to the temporary file as needed
        #         tmp_back.write(b'')

        # # Use the 'tmp_back.name' as the path to the temporary file

        # # copy background fasta
        # tmp_dir = tempfile.TemporaryDirectory()
        # tmp_back_n = tmp_dir.name
        # if args.background != '':
        #     shutil.copy(args.background, tmp_back_n)
        #     tmp_back = os.path.join(tmp_back_n, os.path.basename(args.background))
        # else:
        #     tmp_back = os.path.join(tmp_back_n, "tmp.fasta")
        #     open(tmp_back_n, 'w').close()
        #     tmp_back.write(b'')

        # print(tmp_back)
        # if args.background != '':
        #     with tempfile.TemporaryDirectory() as tmp_back:
        #         # copy the file to the temporary directory
        #         shutil.copy(args.background, tmp_back)
        #         # construct the destination path in the temporary directory
        #         tmp_back = os.path.join(tmp_back, os.path.basename(args.background))
        # else:
        #     # create a temporary file with a .fasta extension
        #     with tempfile.NamedTemporaryFile(suffix='.fasta', delete=False) as tmp_back:
        #         # write an empty string to the file
        #         tmp_back.write(b'')

        # add two random lines for initialization
        with open(tmp_back, "a") as fasta_file:
            fasta_file.write(">appended_rdm_sequence1\n")
            fasta_file.write("gcagcagtcgctcgatccgat" + "\n")
            fasta_file.write(">appended_rdm_sequence2\n")
            fasta_file.write("gcagcagtcggagatagacctcgatccgat")

        # # Read the content of tmp_back and filter out blank lines, in case there were in the input file
        # with open(tmp_back, "r") as fasta_file:
        #     lines = fasta_file.readlines()
        #     filtered_lines = [line.strip() for line in lines if line.strip()]

        # # Write the filtered content back to tmp_back
        # with open(tmp_back, "w") as fasta_file:
        #     fasta_file.write("\n".join(filtered_lines))

        # Call the design_primers function to design primers for the target region.
        primer_tmp = []
        primer_tmp = design_primers(
            input_file=record,
            region_start=start,
            region_end=end,
            target_tm=args.target_tm,
            primer_len=args.primer_len,
            product_size_range=product_size_range,
            name=name,
            ret=args.ret,
            Q5=args.Q5,
            background=tmp_back
            )

        if primer_tmp is not None and len(primer_tmp) > 0:
            primers_all.append(primer_tmp)
        else:
            print("No primer found for: " + name)
        


    #Typically, a ΔG value below -9 kcal/mol is considered to indicate a high likelihood of dimer formation, while values between -6 and -9 kcal/mol indicate a moderate likelihood. Values above -6 kcal/mol are considered unlikely to result in dimer formation.
    # Define initial placeholder score
    best_score = {'size': float('inf'), 'tmp': float('inf'), 'range': float('inf'), 'mean': float('inf')}
    best_comb = []
    # get a set of unique names
    flat_list = []
    for sublist in primers_all:
        for item in sublist:
            flat_list.append(item)

    primers_all = flat_list
    cluster_method = True

    names = set([primer['name'] for primer in primers_all])  # get unique names
    if len(names) <= 1:
        print("not enough targets! " + ' '.join(map(str, names)))
        exit(1)
    else:
        print("Finding best primer set for the following targets: " + ', '.join(map(str, names)))
        if cluster_method:
            best_comb = cluster_primers(primers_all)
        else:
            best_score = {'mean': float('inf'), 'range': float('inf'), 'tmp': float('inf'), 'size': float('inf')}
            best_comb = None
            valid_combinations = []
            pos = 0
            name_primers_dict = {}
            for name in names:
                name_primers_dict[name] = [primer for primer in primers_all if primer['name'] == name]  # group primers by name

            combinations = []
            for i in range(args.eval):
                combination = []
                for name in names:
                    name_primers = name_primers_dict[name]  # get primers for the current name
                    combination.append(random.choice(name_primers))
                combinations.append(combination)

            print("Picking best set from: ", len(combinations), " combinations")
            # tm and size
            product_sizes = [d['Product Size'] for d in primers_all]
            tm_values = [d['Forward tm'] for d in primers_all] + [d['Reverse tm'] for d in primers_all]
            # product_size_range = max(product_sizes) - min(product_sizes)
            tm_range = max(tm_values) - min(tm_values)
            best_score = {'mean': float('inf'), 'range': float('inf'), 'tmp': float('inf')}
            best_comb = None

            with concurrent.futures.ThreadPoolExecutor() as executor:
                futures = [executor.submit(evaluate_combination, comb) for comb in combinations]

                for future in concurrent.futures.as_completed(futures):
                    score, comb = future.result()
                    if best_score['mean'] > score['mean']:
                        if best_score['range'] * 0.9 > score['range']:
                            best_score = score
                            best_comb = comb
                        elif best_score['tmp'] > 4 or score['tmp'] > 4 and best_score['tmp'] > score['tmp']:
                            best_score = score
                            best_comb = comb
            
        ######
        print("")
        print("writing file: " + args.output + '.xlsx')
        # convert best_comb tuple to DataFrame object
        best_comb = pd.DataFrame(list(best_comb))
        # add in NEB calc.
        best_comb['Forward Primer TM NEB'] = best_comb['Forward Primer'].apply(q5_melting_temp)
        best_comb['Reverse Primer TM NEB'] = best_comb['Reverse Primer'].apply(q5_melting_temp)

        best_comb.to_excel(args.output + '.xlsx', sheet_name='PrimerSet', index=False, engine='openpyxl') 


    tmp_dir.cleanup()

# (plt, tab) = plot_cluster(clusterer, primers)
# print(tab)

if __name__ == '__main__':
    main()


#!/bin/python
import os

def main():
    print("Welcome to primerJinn")
    print("")
    os.system('getMultiPrimerSet')
    print("")
    os.system('PCRinSilico')

if __name__ == '__main__':
    main()